
(* SPAN - A Security Protocol Animator
   Copyright (C) 2005-2006 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.0. Last modification 30/11/06
*)
open Types;;
open Globals;;
open Utils;;
open Interface;;
open Utils_debug;;
open Section_rules;;
open Error_handler;;
open Semantic_check;;


(*
  \section[Module Variables]{Definition of global variables in this module}

  \subsection{Variables}
*)


(*
   We define these dummy values in order to simplify the handling of
   composition/basic roles
 \begin{center}
  type: [val dummy_composition_role : Types.composition_role ref]
 \end{center}
*)
let dummy_composition_role = 
  ref (new Types.composition_role (Call (new Types.call_type 0 [])))
;;

(*
  \begin{center}
  type: [val dummy_basic_hlpsl_role : Types.basic_hlpsl_role ref]
 \end{center}
*)
let dummy_basic_hlpsl_role = ref new Types.basic_hlpsl_role ;;


(*
   List of all the hlpsl roles -- basic as well as composition.
  \begin{center}
  type: [val role_def_list : (int * Types.role_type) list ref]
 \end{center}
*)
let role_def_list = ref ([]: (int * Types.role_type) list);;


(*
   List of all the basic role instances prepared during IF generation.
  \begin{center}
  type: [val basic_role_instances : (int * Types.role_instance) list ref]
 \end{center}
*)
let basic_role_instances = ref ([]: (int * role_instance) list);;

(*
   Combination of basic roles instances to consider.
  \begin{center}
  type: [val instances_schema : composition_schema ref]
 \end{center}
*)
let instances_schema = ref (None: composition_schema);;


(*
   List of goals.
  \begin{center}
  type: [val goals : Types.goals_type list ref]
 \end{center}
*)
let goals = ref ([]: goals_type list)

(*
   List of all the roles which are played by the intruder
   generated during the generation of instances.
  \begin{center}
  type: [val intruder_roles : int list ref]
 \end{center}
*)
let intruder_roles = ref ([]: int list);;

(*

  \section{Functions on module's variables}

   We define [find_role] to get a role from [role_def_list] by 
   a given role identifier. 
*)

(*
  \begin{center}
  type:
   [val make_sid_variable : int -> int -> int]
 \end{center}
*)
let make_sid_variable session_id role_id =
  let sid_var_string_id =
    Globals.string_id#register_name
      "SID"
      role_id
  in
  let sid_const_string_id =
    string_id#register_name
      (string_of_int session_id)
      role_id
  in
  let sid_var_type =
    type_table#register_type
      sid_var_string_id
      (Base(Nat))
  in
  if sid_var_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID has to be of type nat","","","","");
  let sid_const_type =
    type_table#register_type 
      sid_const_string_id 
      (Base(Nat)) 
  in
  if sid_const_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined constant "^(string_of_int session_id)^" has to be of type nat","","","","");
 let sid_var_id =
    Globals.global_var_id#get_id_of 
      session_id 
      sid_var_string_id 
  in
  let sid_var_mmap_id =
    Globals.memory_map#map_new_var 
      sid_var_id 
  in
(*  let sid_var_value =
    Globals.value_table#set_value
      sid_var_mmap_id
      (If_const(sid_const_string_id))
  in *) 
  Globals.value_table#set_value
      sid_var_mmap_id
      (If_const(sid_const_string_id));
  sid_var_id
;;

(*
  \begin{center}
  type:
   [val find_role : int -> Types.role_type]
 \end{center}
*)
let find_role some_role =
  try (
    snd
      (List.find
	 (fun (a,_) -> a = some_role)
	 !role_def_list))
  with Not_found ->
    (handleSemanticError (0,"",(Globals.string_id#get_name some_role),"","","");
     raise Not_found)
;;

(*
   The [find_generic_role] has the same purpose, but is simpler to use if one doesn't need
   to know whether the role is [basic_hlpsl_role] or a [composition_role].
  \begin{center}
  type:
   [val find_generic_role : int -> Types.generic_hlpsl_role]
 \end{center}
*)
let find_generic_role role_id =
  match (find_role role_id) with
    Basic x -> (x :> Types.generic_hlpsl_role)
  | Composition x -> (x :> Types.generic_hlpsl_role)
;;


(*  

    \section{Generic Functions}

   Global identifiers are defined from a combinaison of variable
   identifiers and call identifiers. To each role call argument
   corresponds a global identifier.  [register_inst_args] registers these
   global identifiers in the [global_var_id] table.  A new variable is
   created if the argument is not an atomic identifier.
  \begin{center}
  type:
   [val register_list_of_var : int -> int list -> int list]
 \end{center}
*)
let register_list_of_var session_id =
  List.map
    (fun x ->
      global_var_id#get_id_of session_id x)
;;

(* 
   In the IF level, some types, like channels, will not appear in the
   state predicate. We define [filter_args] to remove identifiers of such
   types from a list of arguments. The argument is an association list between
   keys in the [global_var_id] table and terms (in the [term] type).
  \begin{center}
  type:
   [val filter_args : (Types.term_type -> bool) -> (int * 'a) list -> (int * 'a) list]
 \end{center}
*)
let filter_args not_excluded =
  List.find_all 
    (fun (param,_) ->
      let param_type = var_id_to_type param in
(*      let real_type = to_if_type param_type in *)
      not_excluded param_type)
;;


(*
   The "local" variables of a role identified by [role_id] should be registred 
   in the [global_var_id] table. This function performs this registration.
  \begin{center}
  type:
   [val register_inst_local : int -> < get_id : int; .. > list -> int list]
 \end{center}
*)
let register_inst_local session_id vars =
  register_list_of_var session_id
    (List.map (fun var -> var#get_id) vars)
;;


(*
   We map role arguments and parameters global identifiers so that they point 
   to the same location in the memory table.

   CORRECTION: as far as I have understood, the rule is now pass-by-value.
   this function (as well as unnecessary complications) is now obsolete.
  \begin{center}
  type:
   [val new_to_old_mem_map : int list -> int list -> unit list]
 \end{center}
*)
let new_to_old_mem_map =
  List.map2
    (fun param arg ->
      memory_map#map_var_to param arg)
;;


(*
   This utility function replaces the variables of a term by the corresponding 
   constants. It differs from [get_if_value_type] by the return value which is of 
   type term instead of [if_value_term].
  \begin{center}
  type:
   [val replace_vars_by_consts :
   int ->
   Types.atoms_type Types.term_structure ->
   Types.atoms_type Types.term_structure]
 \end{center}
*
let replace_vars_by_consts session_id =
  let rec replace_vars_by_consts_rec = function
      Pair(t1,t2) ->
	Pair(replace_vars_by_consts_rec t1, 
	     replace_vars_by_consts_rec t2)
    | Scrypt(t1,t2) ->
	Scrypt(replace_vars_by_consts_rec t1, 
	       replace_vars_by_consts_rec t2)
    | Crypt(t1,t2) ->
	Crypt(replace_vars_by_consts_rec t1,
	      replace_vars_by_consts_rec t2)
    | Inv(t) ->
	Inv(replace_vars_by_consts_rec t)
    | Function(t1,l) ->
	Function(replace_vars_by_consts_rec t1 ,
		 List.map replace_vars_by_consts_rec l)
    | Cons(t1,t2) -> 
	Cons(replace_vars_by_consts_rec t1, 
	     replace_vars_by_consts_rec t2)
    | Set(l) ->
	Set(List.map replace_vars_by_consts_rec l)
    | Lst(l) ->
	Lst(List.map replace_vars_by_consts_rec l)
    | Base(Var(i)) -> 
	begin
	  let if_value = 
	    value_table#get_value
	      (memory_map#get_map
		 (global_var_id#get_id_of session_id i))
	  in
	  match if_value with 
	    If_const c -> Base(Const(c))
	  | _ -> Base(Var(i))
        end
    | Base(Const(i)) as t -> t
    | t -> displayWarning 3 "replace_vars_by_consts"; t
  in
  replace_vars_by_consts_rec
;;
*)


(*
   [played_by_intruder] computes if a given role is played by the intruder.
  \begin{center}
  type:
   [val played_by_intruder : < get_player : int; .. > -> int -> bool]
 \end{center}
*)
let played_by_intruder basic_role session_id =
  let player = basic_role#get_player in
  let player_instance =
    Globals.value_table#get_value 
      (Globals.memory_map#get_map
	 (Globals.global_var_id#get_id_of session_id player))
  in
  if (!debug_level >= 2) then
    begin
      prerr_string "  instance of player: "; 
      Interface.prerr_ground_if_value player_instance; 
      prerr_newline(); 
    end;
  match player_instance with
    If_const(n) -> (n = Globals.global_i)
  | _ -> false
;;


(*
   The next function does the same task, but takes care, for sets, to
   create a constant of type pointer that will hold the value. The
   [mmap_id] is the key in the [memory_map] table for a variable of
   type [set] in the parameters of the role and the [if_value] is
   supposed to be a set. These conditions have to be verified in the
   [assign_argument] function before making a call to the
   [assign_set_argument] function. In case of policy change, we hope
   to be able to re-use this function.
  \begin{center}
  type:
   [val assign_set_value : int -> Types.if_value_type -> unit]
 \end{center}
*)
let assign_set_value elts =
  (*
    First, we create a new string that will be used as a pointer to the set.
    We register its type as [Pointer], register it in the [global_var_id] table, 
    give it a value ni the [memory_map] table. Informally, this value is the
    address of this set in the memory.
  *)
  let pointer_string_id = string_id#register_blind_const "set" 0 in
  let pointer_global_var_id = Globals.global_var_id#get_id_of 0 pointer_string_id in
  let pointer_mmap_id = Globals.memory_map#map_new_var pointer_global_var_id in

	let eltType =
	  let rec find_applied = function
	      t::lt ->
		(try
		  [Utils.if_term_to_type t]
		with Not_found -> find_applied lt)
	    | [] -> []
	  in
	  find_applied elts
	in
(*  ignore (Globals.type_table#register_type pointer_string_id (Base(Pointer)));*)
(*  ignore (Globals.type_table#register_type pointer_string_id (Set([Base(Message)])));*)
  ignore (Globals.type_table#register_type pointer_string_id (Set(eltType)));
  Globals.value_table#set_value pointer_mmap_id (If_set(elts));
  if (!debug_level >= 3) then
    prerr_endline ("  creating pointer "^(Globals.string_id#get_name pointer_string_id)
		   ^" for value "^(ground_if_value_to_string (If_set(elts))));
  If_const(pointer_string_id)
;;


(*
  \begin{center}
  type:
   [val assign_value :
   int -> int -> Types.atoms_type Types.term_structure -> unit]
 \end{center}
*)
let assign_value rhs_session_id global_id term =
  
  (*
    The [get_if_value_type] function may raise an exception if the variables
    in the term do not already have a value in the [value_table]
  *)
  let if_value_term = get_if_value_type rhs_session_id term in
  let real_if_value_term =
    match if_value_term with
(*      If_lst(l) -> assign_set_value l*)
      If_set(l) -> assign_set_value l
    | _ -> if_value_term
  in
  Globals.value_table#set_value (memory_map#get_map global_id) real_if_value_term
;;


(*
   The function [compute_knowledge_given_to_intruder] computes the knowledge
   given to the intruder \emph{when he is playing a role}.
  \begin{center}
  type: [val compute_knowledge_given_to_intruder :
   generic_hlpsl_role -> int -> unit]
 \end{center}
-----------------------
let compute_knowledge_given_to_intruder role_found session_id =
  let agents_knowledge_list = role_found#get_knowledge in
  List.iter
    (fun (agent, knowledge) ->
      match (get_if_value_type session_id agent) with
	If_const(i) ->
	  if (is_intruder i) then
	    let knowledge_list =
	      List.map
		(fun x ->
		  let if_value_term = get_if_value_type session_id x in
		  match if_value_term with
		    If_lst(l) ->
		      assign_set_value l
		  | If_set(l) ->
		      assign_set_value l
		  | _ -> 
		      if_value_term)
		knowledge
	    in
	    i_knowledge := concatenate_with_no_copy knowledge_list !i_knowledge
      | _ -> displayWarning 1 "Instantiation.compute_knowledge_given_to_intruder")
    agents_knowledge_list
*)
let compute_knowledge_given_to_intruder role_found session_id =
  let knowledge_list =
    List.map
      (fun x ->
	let if_value_term = get_if_value_type session_id x in
	match if_value_term with
(*	  If_lst(l) ->
	    assign_set_value l*)
	  If_set(l) ->
	    assign_set_value l
	| _ -> 
	    if_value_term)
      (role_found#get_knowledge)
  in
  i_knowledge := concatenate_with_no_copy knowledge_list !i_knowledge
;;


(*
   The [compute_assignment_list] function takes as input: 
   \begin{itemize}
   \item an [(x,(y,z)) list] where [x] is a reference to a boolean, [y]
   is a variable identifier in the [global_var_id] table and [z] is a term
   representing the Right-Hand Side of the assignment in the [term]
   type;
   \item an integer [RHS_session_id] representing the session id of the Right-Hand side terms.
   \end{itemize}
  \begin{center}
  type:
   [val compute_assignment_list :
   int ->
   ('a * (int * Types.atoms_type Types.term_structure)) list ->
   (bool * (int * Types.atoms_type Types.term_structure)) list]
 \end{center}
*)
let compute_assignment_list rhs_session_id =
  let assignment = assign_value rhs_session_id in
  let compute_assignment (_,(y,z)) =
    try
	assignment y z;
	(false,(y,z))
    with
      Not_found ->
	(true,(y,z))
    | Variable_has_no_mmap x ->
	(true,(y,z))
  in
  List.map compute_assignment
;;


(*
   The [compute_assignation_list] function computes an unordered list of assignation.
   It is used in the [init] section of a role declaration and when a role call is done.
   The list of assignation is specified by two lists and two session identifiers.
   The parameters are:
   \begin{itemize}
   \item [rhs_term_value_list]: The list of terms (type [term]) in the right-hand side of assignation.
   These terms can be any term. The [in] predicate is not currently handled;
   \item [lhs_variables_string_id]: The list of variables in the left-hand of assignations.
   In the future, one could put some functions here. The list is a list of valid string identifiers 
   in the [Globals.string_id] variable;
   \item [lhs_session_id]: This session identifier, taken from [role_instance], permits to assign
   global variables, if necessary, to variables on the left-hand side of equations;
   \item [rhs_session_id]: This session identifier permits to retrieve the value of the terms in the
   right-hand side of an assignation. 
   \end{itemize}
   The return value is an association list containing, for each argument, the key
   in the [global_var_id] table of its variable in the role instance linked
   with the parameter of the call (of type [term]).
  \begin{center}
  type:
   [val compute_assignation_list :
   Types.atoms_type Types.term_structure list ->
   int list ->
   int -> int -> (int * Types.atoms_type Types.term_structure) list]
 \end{center}
*)
let compute_assignation_list
  rhs_term_value_list lhs_variables_string_id
  lhs_session_id rhs_session_id =

  if (!debug_level >= 2) then
    (prerr_string "% compute_assignation_list:  lhs_session_id=";
     prerr_int lhs_session_id;
     prerr_string "  rhs_session_id=";
     prerr_int rhs_session_id; prerr_newline();
     prerr_string "  variables:");
  let lhs_global_var_id = 
    List.map
      (fun x -> 
	if (!debug_level >= 2) then
	  prerr_string ("  "^(Globals.string_id#get_name x));
	Globals.global_var_id#get_id_of lhs_session_id x)
      lhs_variables_string_id
  in
  if (!debug_level >= 2) then
    (prerr_string "\n     values:";
     List.iter 
       (fun x -> prerr_string "  "; Interface.prerr_term x)
       rhs_term_value_list;
     prerr_newline());

  (*
     For the assignation, we are computing a fix-point..
   *)
  let assignment_list =
    List.combine lhs_global_var_id rhs_term_value_list
  in

  (*
     We want to use the next function as less as possible. It permits to register variables
     in the [memory_map] table.
   *)

  (*
     We are doing a fix point computation. In the worst case, the number of iterations is 
     the length of the list.
   *)
  let length = List.length assignment_list in
  let i = ref 0 in
  let was_not_assigned (x,_) = x in
  let complete_assignment_list = 
    List.map (fun x -> (true,x)) assignment_list
  in
  let not_already_assigned_list = ref complete_assignment_list in

  (*
     First, we register once and for all the variables in the [memory_map] table.
   *)
  List.iter memory_map#check_map_var lhs_global_var_id;

  (*
     Second, we compute the fix point of the assignation list.
   *)
  (* version with sem. error when variable not instantiated:
  (try
      while (!not_already_assigned_list <> []) do
	not_already_assigned_list := 
	  List.find_all 
	    was_not_assigned
	    (compute_assignment_list
	       rhs_session_id 
	       !not_already_assigned_list);
	incr i;
        if (!i > length) then 
          raise Not_found
      done
    with Not_found ->
      (handleSemanticError
	 (20,
	  "Semantical error: impossible instantiation.\n"
	  ^"      The following variables have no value: "
	  ^(list_to_string (fun (_,(_,t)) -> term_to_string t) ", " !not_already_assigned_list),
	  "","","","");
       if (!debug_level >= 5) then
	 Utils_debug.dump ();
       exit 0));
   *)
   while ((!not_already_assigned_list <> []) && (!i <= length)) do
     not_already_assigned_list := 
       List.find_all 
         was_not_assigned
         (compute_assignment_list
            rhs_session_id 
            !not_already_assigned_list);
     incr i;
   done;
   if (!i > length) then
     (not_already_assigned_list :=
       List.filter
	 (fun (_,(x,_)) ->
	   not_excluded_type
	     (Globals.type_table#get_type (snd (Globals.global_var_id#get_variable x))))
	 !not_already_assigned_list;
      if (!not_already_assigned_list <> []) then 
	displayWarning 10
	  ("incomplete instantiation of a role call.\n"
	   ^"%%    The following variables have no value: "
           ^(list_to_string
	       (fun (_,(_,t)) ->
		 term_to_string t)
	       ", " !not_already_assigned_list)));

  (*
     In the return value, we just give the association list between a key in the [global_var_id]
     table and its value in the [term] type. 
   *)
  if (!debug_level >= 3) then
    Utils_debug.affiche_debug assignment_list;
  assignment_list    
;;


(*
  \begin{center}
  type:
   [val extract_init_infos : Types.predicate list -> (int list * Types.term list * Types.predicate list)]
 \end{center}
*)
let extract_init_infos =
  let rec extract_rec = function
      Equal(Base(Var(l)),r)::lpred ->
	let (ll,lr,lp) = extract_rec lpred in
	(l::ll,r::lr,lp)
    | Equal(_)::lpred ->
	let (ll,lr,lp) = extract_rec lpred in
	(ll,lr,lp)
    | (Event(e,l))::lpred ->
	let (ll,lr,lp) = extract_rec lpred in
	(ll,lr,(Base(e),l)::lp)
    | _::lpred ->
	displayWarning 0 "Warning: init section: predicate not supported";
	extract_rec lpred
    | [] -> ([],[],[])
  in
  extract_rec
;;


(*
  \begin{center}
  type:
   [val compute_initial_infos : generic_hlpsl_role -> int -> int list]
 \end{center}
*)
let compute_initial_infos role_found session_id =
  let local_variables =
    (* to remove all variables with excluded type *)
    List.filter
      (fun v ->
	Globals.not_excluded_type v#get_type)
      role_found#get_local
  and (init_infos_lhs_var_id, init_infos_rhs, init_preds) =
    extract_init_infos role_found#get_init
  in

  ignore (compute_assignation_list init_infos_rhs init_infos_lhs_var_id session_id session_id);

  (*
     The important part, here, is the [compute_gid_of_var]. The [unsafe_get_id_of] method permits
     to know which variables do not have an initial value. These variables are considered as nonces.
   *)
  let compute_gid_of_var x =
    try
      if (!debug_level >= 3) then
	(prerr_string "  identifier of the variable: "; 
	 prerr_int x#get_id; 
	 prerr_newline ());
      Globals.global_var_id#unsafe_get_id_of session_id x#get_id 
    with
	_ ->
(
(*** essai pour conserver les noms de variables qd non instancies ***)
	  let x_type = Globals.type_table#get_type x#get_id in
	  let dummy_value = dummy_val_of_type x_type in
	  ignore (compute_assignation_list 
		    [dummy_value] 
		    [x#get_id]  
		    session_id session_id); (***)
	  if (!debug_level >= 3) then
	    (prerr_endline("  assignment of a dummy value "^
                           (term_to_string dummy_value)^" to "
			   ^(Globals.string_id#get_name x#get_id)));
(*** essai pour conserver les noms de variables qd non instancies ***)
	  Globals.global_var_id#unsafe_get_id_of session_id x#get_id (***)
(**	  ignore (compute_assignation_list
		    [Base(Var(x#get_id))] 
		    [x#get_id]  
		    session_id session_id);
          Globals.global_var_id#unsafe_get_id_of session_id x#get_id **)
 )
  in

  init_predicates :=
    List.fold_left
      (fun ip (p,lt) ->
	try
	  insert_with_no_copy
	    (get_if_value_type session_id p,
	     List.map (get_if_value_type session_id) lt)
	    ip
	with _ ->
	  displayWarning 10
	    ("Role "^(string_id#get_name role_found#get_name)^"\n"
             ^"%%    Cannot instantiate \""
             ^(if_term_to_string "" (Unknown(p,lt)))
	     ^"\" in init section... dropped!");
	  ip)
      !init_predicates init_preds;

  (* Extracting valid gids *)
  List.find_all
    (fun x -> x <> 0)
    (List.map compute_gid_of_var local_variables)
;;


(* Removed on April 6, 2005 (only intruder knowledge)
  \begin{center}
  type:
   [val compute_additional_knowledge : basic_hlpsl_role -> int -> int list]
 \end{center}
*
let compute_additional_knowledge basic_role session_id =
  let (all_knlge_gids,new_knlge_sids,new_knlge_val) =
    * all_knlge_gids: list of the variable id of each knowledge
       new_knlge_sids: list of the string id of each knowledge (<> variable)
       new_knlge_val: list of the term value of each knowledge (<> variable) *
    let role_id = basic_role#get_name
    and knlge_infos = basic_role#get_knowledge
    and player_id = basic_role#get_player in
    List.fold_left  * boucle sur les knowledge(xa) *
	 (fun (lgid,lvar,lval) (agent,knowledge) ->
	   if (match agent with
	         Base(Var(xa)) -> (xa = player_id)
               | Base(Const(ca)) -> (get_if_value_type session_id (Base(Var(player_id))) = If_const(ca))
	       | _ -> false) then
	     List.fold_left  * loop over the listed knowledge *
	       (fun (lgid,lvar,lval) xv ->
		 match xv with
		   Base(Var(xid)) ->
		     if Globals.not_excluded_type (Globals.type_table#get_type xid) then
		       (lgid@[Globals.global_var_id#get_id_of session_id xid], lvar, lval)
		     else
		       (lgid,lvar,lval)
	         | Base(Const(xid)) ->
		     let cst_type = Globals.type_table#register_type xid (Base(Message)) in
		     if Globals.not_excluded_type cst_type then
		       (let new_string_id = Globals.string_id#register_blind_var "Knlg" role_id in
		        let new_var_id = Globals.global_var_id#get_id_of session_id new_string_id in
			let new_mmap_id = Globals.memory_map#map_new_var new_var_id in
			ignore (Globals.type_table#register_type new_string_id cst_type);
			Globals.value_table#set_value new_mmap_id (get_if_value_type session_id xv);
			(lgid@[new_var_id],lvar@[new_string_id],lval@[xv]))
		     else
		       (lgid,lvar,lval)
		 | Inv(Base(Var(xid))) ->
		     let new_string_id = Globals.string_id#register_blind_var "Knlg" role_id in
		     let new_var_id = Globals.global_var_id#get_id_of session_id new_string_id in
		     let new_mmap_id = Globals.memory_map#map_new_var new_var_id in
		     ignore (Globals.type_table#register_type new_string_id (Inv(Base(Public_key))));
		     Globals.value_table#set_value new_mmap_id (get_if_value_type session_id xv);
		     (lgid@[new_var_id],lvar@[new_string_id],lval@[xv])
	         | Inv(Base(Const(xid))) ->
		     let cst_type = Globals.type_table#register_type xid (Base(Public_key)) in
		     (if cst_type <> (Base(Public_key)) then
		       let error_string =
			 ("constant "^(Globals.string_id#get_name xid)^
			  " used with incompatible type in knowledge declaration of role"
			  ^(Globals.string_id#get_name player_id))
		       in
		       Semantic_check.add_sem_error
			 (20, error_string,"","","",""));
		     let new_string_id = Globals.string_id#register_blind_var "Knlg" role_id in
		     let new_var_id = Globals.global_var_id#get_id_of session_id new_string_id in
		     let new_mmap_id = Globals.memory_map#map_new_var new_var_id in
		     ignore (Globals.type_table#register_type new_string_id (Inv(Base(Public_key))));
		     Globals.value_table#set_value new_mmap_id (get_if_value_type session_id xv);
		     (lgid@[new_var_id],lvar@[new_string_id],lval@[xv])
		 | _ ->
		     let xv_type = Utils.term_to_type xv in
		     if Globals.not_excluded_type xv_type then
		       (let new_string_id = Globals.string_id#register_blind_var "Knlg" role_id in
		        let new_var_id = Globals.global_var_id#get_id_of session_id new_string_id in
			let new_mmap_id = Globals.memory_map#map_new_var new_var_id in
			ignore (Globals.type_table#register_type new_string_id xv_type); * was: Base(Message) *
			Globals.value_table#set_value new_mmap_id (get_if_value_type session_id xv);
			(lgid@[new_var_id],lvar@[new_string_id],lval@[xv]))
		     else
		       (lgid,lvar,lval))
	       (lgid,lvar,lval) knowledge
	   else
	     (lgid,lvar,lval))
	 ([],[],[]) knlge_infos
  in
  * to assign the value of each newly generated variable *
  ignore (compute_assignation_list new_knlge_val new_knlge_sids session_id session_id);
  all_knlge_gids
;;
*)

(*
   [unfold_instances] is the main instantiation function. It builds recursively
   the role instances. It is first called from [call_instantiation] with main call 
   as argument. For each basic role call, we associate a role instance. This function
   uses the [section_init] module to built the initial state of each basic role.
    It also uses [section_rules] module to build the IF rules. Similar basic role
   instance have the same IF rules. Later steps of the execution will print these
   rules only once. However, we keep the rules in all the instances so that we
   may easily instantiate them if we need it in future versions of the compiler.
  \begin{center}
  type:
   [val get_set_if_value : int -> int -> Types.if_value_type]
 \end{center}
*)
let get_set_if_value session_id set_var =
  let set_var_term = (get_if_value_type session_id  (Base(Var(set_var)))) in
  let set_pointer_string_id =
    match set_var_term with
      If_const (x) -> x
    | _  ->
	(displayWarning 0 
	   ("Instantiation.get_set_if_value: trying to get the value of an uninitialized set: "
	    ^(Interface.term_to_string (Base(Var(set_var))))
	    ^" = "
	    ^(Interface.ground_if_value_to_string set_var_term));
	 raise Not_found)
  in
  let set_pointer_gid = global_var_id#get_id_of 0 set_pointer_string_id in
  let set_pointer_mmap_id = memory_map#get_map set_pointer_gid in
  value_table#get_value set_pointer_mmap_id
;;


(*TEST*)
let rec ifpair_to_list = function
    If_pair(a,b) -> a::(ifpair_to_list b)
  | c -> [c]
;;


(*
  \begin{center}
  type:
   [val unfold_instances : Types.composition_type -> Types.composition_type]
 \end{center}
*)
let unfold_instances main_call =

let basic_roles_encontered = ref [] in

let rec unfold_instances_rec some_composition_type prev_role_instance =

  let handle_multi_call composition_over composition_expression =

    let session_id = prev_role_instance#get_session_Id in
    let set_of_composition = composition_over#get_set in
    let set_value = get_set_if_value session_id set_of_composition in

    let set_of_composition_instance = 
      match set_value with
(*	If_lst(l)*)
	If_set(l) ->
	  List.map 
	    (function
(*		If_lst u  -> 
		  if (!debug_level >= 2) then
		    (prerr_string "  it contains the term "; 
		     Interface.prerr_ground_if_value (If_lst u); 
		     prerr_newline());
		  u*)
	      | If_set u -> 
		  if (!debug_level >= 2) then
		    (prerr_string "  it contains the term "; 
		     Interface.prerr_ground_if_value (If_set u); 
		     prerr_newline());
		  u
	      | If_pair(_) as p ->  (*TEST*)
		  ifpair_to_list p
	      | wrong -> 
		  (displayWarning 0
		     ("Instantiations.unfold_instances: a set used for composition does not contain lists, it contains the term "
		      ^(Interface.ground_if_value_to_string wrong));
		   raise Not_found))
	    l

      | _ -> 
	  displayWarning 0 "Instantiations.unfold_instances: value pointed to by the set pointer is not a set"; 
	  raise Not_found
    in
    let variable_list_in_term_type =
      pair_to_list composition_over#get_term
    in

    let variable_list_in_global_id =
      List.map
	(variable_in_term_type_to_variable_global_id session_id)
	variable_list_in_term_type
    in

    if set_of_composition_instance = [] then
      (displayWarning 0 "Instantiations.unfold_instances: empty set/list in parallel composition!";
       None)
    else
      let treat_one_instance instance =
	List.iter2
	  (fun global_id if_value_term ->
	    ignore
	      (Globals.value_table#set_value 
		 (memory_map#get_map global_id) 
	         if_value_term))
	  variable_list_in_global_id instance;
	let t = unfold_instances_rec composition_expression prev_role_instance in
	t
      in
      let lt =
	List.fold_left
	  (fun lt instance ->
	    (match treat_one_instance instance with
	      None -> lt
	    | Par(l) -> lt@l
	    | t -> lt@[t]))
	  [] set_of_composition_instance
      in
      if List.length lt = 0 then None
      else if List.length lt = 1 then List.hd lt
      else Par(lt)
  in

  match some_composition_type with
	
    (*
      There is currently no handling of sequential composition, either of two processes 
      or as a loop. It is treated as parallel composition.
    *)
    
    MultiPar (multi_call,composition) ->
      if (!debug_level >= 2) then
	prerr_endline "%% MultiPar...";
      let t = handle_multi_call multi_call composition in
      if (!debug_level >= 2) then
	prerr_endline "%% End MultiPar...";
      t

  | Sequential(type1,type2) ->
      displayWarning 1 "Instantiation.unfold_instances: sequential composition";
      if (!debug_level >= 2) then
	prerr_endline "%% Sequential...";
      let t1 = unfold_instances_rec type1 prev_role_instance in
      let t2 = unfold_instances_rec type2 prev_role_instance in
      if (!debug_level >= 2) then
	prerr_endline "%% End Sequential...";
      (match t1,t2 with
	_,None -> t1
      | None,_ -> t2
      | Seq(l1),Seq(l2) -> Seq(l1@l2)
      | Seq(l1),_ -> Seq(l1@[t2])
      | _,Seq(l2) -> Seq(t1::l2)
      | _ -> Seq([t1;t2]))

  | Parallel(type1,type2) ->
      if (!debug_level >= 2) then
	prerr_endline "%% Parallel...";
      let t1 = unfold_instances_rec type1 prev_role_instance in
      let t2 = unfold_instances_rec type2 prev_role_instance in
      if (!debug_level >= 2) then
	prerr_endline "%% End Parallel...";
      (match t1,t2 with
	_,None -> t1
      | None,_ -> t2
      | Par(l1),Par(l2) -> Par(l1@l2)
      | Par(l1),_ -> Par(l1@[t2])
      | _,Par(l2) -> Par(t1::l2)
      | _ -> Par([t1;t2]))

  | Call(calltyp) ->
      (if (!debug_level >= 2) then
	prerr_string "%%% Call...";  
       (*
	  First, we retrieve all the relevant information directly available
	  from the [calltyp] object.
	*)
	let roleinstance = (new Types.role_instance calltyp) in
	let session_id = roleinstance#get_session_Id in
	if (!debug_level >= 2) then
	  prerr_string ((string_of_int session_id)^"..."^(string_of_int calltyp#get_role)^"...");
	let prev_session_id = prev_role_instance#get_session_Id in
	let role_id = calltyp#get_role in
	let (is_basic_role,basic_role,comp_role,role_found) = 
	  match (find_role role_id) with
	    Basic br -> 
	      (true,br,!dummy_composition_role,(br :> Types.generic_hlpsl_role))
	  | Composition br ->
	      (false,!dummy_basic_hlpsl_role,br,(br :> Types.generic_hlpsl_role))
	in
	if (!debug_level >= 2) then
	  prerr_endline (Globals.string_id#get_name (role_found#get_name));

	(*
	   First, we register all the variables in the argument of the
	   role, and give them a value according to the call of the
	   role. The return value is an association list between key
	   of these variables in the [global_var_id] table and the
	   list of terms (in the [term] type) in the call of the role.
	*)
	let all_param =
	  compute_assignation_list
	    (calltyp#get_args)
	    (List.map (fun x -> x#get_id) role_found#get_param)
	    session_id
	    prev_session_id
	in

        (*
	   We remove from the [all_param] association list the excluded types (e.g. channels).
	 *)
	let effective_param =
	  filter_args
	    (if is_basic_role then
	       not_excluded_type
	     else (fun _ -> true))
	    all_param
	in
	(*
	   In the sequel, we need the list of actual arguments when the 
	   role is called or the list of keys in the [global_var_id] table.
	 *)
	let (param_gids,call_args) = 
	  List.split effective_param 
	in
	(*
	   The [local_gids] is the list of keys in the [global_var_id] table
	   corresponding to variables declared in the [local] part of 
	   a role specification.
	 *)
	let local_gids =
	  compute_initial_infos 
	    role_found
	    session_id 
	in

	(*
	   The next function computes the knowledge given to the intruder through
	   the [intruder_knowledge] section of an HLPSL role specification.
	 *)
	compute_knowledge_given_to_intruder role_found session_id;

	(*
	   Last, we compute either the initial state, for the basic roles,
	   or prepare the composition, for composition roles.
	 *)

	if is_basic_role then
	  (
	   (*
	      The next function computes the knowledge given to the intruder through
	      the [knowledge] section of an HLPSL role specification.
	    *
	   let new_knldge_gids = compute_additional_knowledge basic_role session_id in*)

	   (*
	      We set the [variable_instance] list in the role, which will later be used
	      to determine the state of the role in the IF output.
	   roleinstance#set_variable_instances (param_gids @ new_knldge_gids @ local_gids);
	    *)
	   roleinstance#set_variable_instances (param_gids @ local_gids);

	   (*
	      It is stated that the arguments of a role are implicitely known by
	      this role. Thus, if a role is played by the intruder, we add these arguments
	      to its initial knowledge.
	    *)
	   if (played_by_intruder basic_role session_id) then
	     (if (!debug_level >= 2) then 
	       prerr_endline ("% Role played by the intruder: "
			      ^(string_id#get_name role_found#get_name)))
(*	      List.iter 
		(fun arg ->
		  i_knowledge :=
		    insert_with_no_copy
		      (match (get_if_value_type prev_session_id arg) with
			If_lst(l) ->
			  assign_set_value l
		      | If_set(l) ->
			  assign_set_value l
		      | arg_value -> arg_value)
		      !i_knowledge)
		call_args)
*)
	   else
	     (if (!debug_level >= 2) then
	       prerr_endline ("% Role not played by the intruder: "
			      ^(string_id#get_name role_found#get_name));

	      if (!debug_level >= 2) then
		prerr_string "% Creation of the SID variable...";
	      let sid = make_sid_variable session_id role_found#get_name in
	      roleinstance#set_variable_instances
		(Utils.remove_duplicate
		   (((Globals.global_var_id#get_id_of 
			session_id 
			(basic_role#get_player))
		     ::(roleinstance#get_variable_instances))
		    @[sid]));
	      if (!debug_level >= 2) then
		prerr_endline " ok.";

	      if (!debug_level >= 2) then
		prerr_string "% Creation of the initial state...";
	      let init_state = 
		[ State((string_id#get_name (roleinstance#get_name)),
			(roleinstance#get_session_Id),
			List.map 
			  (fun x -> Base(Var(x))) 
			  roleinstance#get_variable_instances)
		]
	      in
	      if (!debug_level >= 2) then
		prerr_endline " ok.";

	      (*** WARNING: done for each instance! instead of just once... ***)
	      if (!debug_level >= 2) then
		prerr_endline "% Creation of the (re)actions list...";
	      let actions_list =
		if List.mem (basic_role#get_name) !basic_roles_encontered then
		  (if (!debug_level >= 2) then
		    prerr_endline "% -> already encountered.";
		   [])
		else
		  get_if_reactions 
		    (basic_role#get_name)
		    (List.map snd basic_role#get_actions)
		    (roleinstance#get_variable_instances)
		    session_id
		    !goals
	      in
	      if (!debug_level >= 2) then
		prerr_endline "% -> (re)actions list ok.";

	      if (!debug_level >= 2) then
		prerr_string "% Setting the global variables...";
	      roleinstance#set_initial_state init_state;
	      roleinstance#set_actions actions_list;
	      basic_role_instances := 
		!basic_role_instances @ [(session_id,roleinstance)];
	      if (!debug_level >= 2) then
		prerr_endline " ok.";

	      if (!debug_level >= 2) then
		prerr_endline 
		  ("%%% End of treatment of the basic role "^
		   (Globals.string_id#get_name (role_found#get_name))));
	   Inst(session_id))

	else
	  (
	   (*
	      We set the [variable_instance] list in the role, which will later be used
	      to determine the state of the role in the IF output.
	      -> useless???
	    *)
	   roleinstance#set_variable_instances (param_gids @ local_gids);

	   (*
	      Note that in the case of a composition, the [prev_role_instance] corresponds
	      to the role (type [role_instance]) where the composition takes place !
	    *)
	   if (!debug_level >= 2) then
		prerr_endline 
		  ("%%% End of treatment of composition role "^
		   (string_id#get_name role_found#get_name));
	   unfold_instances_rec (comp_role#get_composition) roleinstance))

  | Empty -> None

in
unfold_instances_rec main_call (new role_instance (new Types.call_type 0 []))
;;


(*
  \begin{center}
  type:
   [val string_of_instances : composition_schema -> string]
 \end{center}
*)
let rec string_of_instances = function
    Seq(lt) ->
      "("^(list_to_string string_of_instances ";" lt)^")"
  | Par(lt) ->
      "("^(list_to_string string_of_instances "|" lt)^")"
  | Inst(no) ->
      string_of_int no
  | _ -> ""
;;


(*
   This is the main function of instantantiation.
   It calls [prepare_instances] to generate the basic role instances
   from the given specification.
  \begin{center}
  type:
   [val call_instantiation :
   specification_class -> Types.role_instance list]
 \end{center}
*)
let call_instantiation specif =

  let main_role = specif#get_main_role in

  (try
    (role_def_list :=
      List.map
	(function
	    Basic(y) as x -> (y#get_name,x)
          | Composition(y) as x -> (y#get_name,x))
	specif#get_roles;
     goals := specif#get_goals;
     instances_schema := unfold_instances main_role;
     if !debug_level >= 2 then
       (prerr_endline ("%% Schema of instances: "
		       ^(string_of_instances !instances_schema)));
     if !debug_level >= 5 then
       Utils_debug.dump());
  with x ->
    (handleSemanticError
       (20,"Instantiation.call_instantiation: An error occurred during the computation in the Instantiation module",
	"","","","");
     Utils_debug.dump();
     raise x));

  List.map snd !basic_role_instances;
;;
